#!/usr/bin/perl -w
#
# 14all.cgi
#
# create html pages and graphics with rrdtool for mrtg + rrdtool
#
# (c) 1999,2000 Rainer.Bawidamann@informatik.uni-ulm.de
#
# use freely, but: NO WARRANTY - USE AT YOUR OWN RISK!

# if RRDs (rrdtool perl module) is not in the module search path (@INC)
# uncomment the following line and change the path appropriatly:
#use lib '/home/rb1/lib/perl5';

# RCS History - removed as it's available on the web

my $rcsid = '$Id: 14all.cgi,v 1.7 2000/02/25 09:03:16 rb1 Exp $';
my $subversion = (split(/ /,$rcsid))[2];
$subversion =~ m/^\d+\.(\d+)$/;
my $version = "14all.cgi 1.0p$1";

my $DEBUG = 0;  # set to 1 to resolve problems with the graph generation

use strict;
use CGI;
use RRDs 1.00008;

sub read_mrtg_config();
sub print_error(@);
sub intmax(@);
sub yesorno($);
sub get_graph_params($$);
sub getworkdir($);
sub getpngdir($);
sub getpngsize($);
sub errorpng();
sub senderrorpng($$$);

my ($q, %targets, @sorted, %config, $cfgfile, $cfgfiledir, %options);
my ($cgidir, @author, @style);

### where the mrtg.cfg file is
# anywhere in the filespace
#$cfgfile = '/home/mrtg/mrtg.cfg';
# relative to the script
#$cfgfile = 'mrtg.cfg';
# use this so 14all.cgi gets the cfgfile name from the script name 
# (14all.cgi -> 14all.cfg)
$cfgfile = '';

# if you want to store your config files in a different place than your cgis:
$cfgfiledir = '';

### cusotmize the html pages
@author = ( -author => 'Rainer.Bawidamann@informatik.uni-ulm.de');
# one possibility to enable stylesheets (second is to use "AddHead[_]:..." in mrtg.cfg)
#@style = ( -style => { -src => 'general.css' });
###

# initialize CGI
$q = new CGI;
$q->import_names('CGI');

# look for the config file
my $meurl = $q->url();
if (defined $CGI::cfg) {
	$cfgfile = $CGI::cfg;
	$cfgfile = "$cfgfiledir/$cfgfile" unless -r $cfgfile;
	print_error("Cannot find the given config file: \<tt>$CGI::cfg\</tt>")
		unless -r $cfgfile;
} elsif (!$cfgfile) {
	$meurl =~ m|/([^/]*)\.cgi$|;
	$cfgfile = $1 . '.cfg';
	$cfgfile = "$cfgfiledir/$cfgfile" unless -r $cfgfile;
}

# read the config file
read_mrtg_config();
# fix some settings
#   make sure icondir ends with /
$config{icondir} .= "/" if $config{icondir} !~ m|/$|;
#   does the configured workdir exist?
#   (if no workdir is given we can use the actual dir)
if ($config{workdir} && ! -d $config{workdir}) {
	print_error($q->p("Cannot access the configured working directory, please check if the line"),
		$q->p($q->tt("Workdir: $config{workdir}")),
		$q->p("(in ",$q->tt($cfgfile),") is typed correctly."),
		$q->p("Remember: this is a path into the global filesystem, not a web server path"));
	exit 0;
}

my @headeropts = ( @author, @style, -bgcolor => $config{background} );

# the footer we print on every page
my $footer = <<"EOT";
<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=0>
  <TR>
    <TD WIDTH=63><A ALT="MRTG"
    HREF="http://ee-staff.ethz.ch/~oetiker/webtools/mrtg/mrtg.html"><IMG
    BORDER=0 SRC="$config{icondir}mrtg-l.gif"></A></TD>
    <TD WIDTH=25><A ALT=""
    HREF="http://ee-staff.ethz.ch/~oetiker/webtools/mrtg/mrtg.html"><IMG
    BORDER=0 SRC="$config{icondir}mrtg-m.gif"></A></TD>
    <TD WIDTH=388><A ALT=""
    HREF="http://ee-staff.ethz.ch/~oetiker/webtools/mrtg/mrtg.html"><IMG
    BORDER=0 SRC="$config{icondir}mrtg-r.gif"></A></TD>
  </TR>
</TABLE>
<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=0>
  <TR VALIGN=top>
  <TD WIDTH=88 ALIGN=RIGHT><FONT FACE="Arial,Helvetica" SIZE=2>Version 2.8.x-rrd</FONT></TD>
  <TD WIDTH=388 ALIGN=RIGHT><FONT FACE="Arial,Helvetica" SIZE=2>
  <A HREF="http://ee-staff.ethz.ch/~oetiker/">Tobias Oetiker</A>
  <A HREF="mailto:oetiker\@ee.ethz.ch">&lt;oetiker\@ee.ethz.ch&gt;</A> and
  <A HREF="http://www.bungi.com">Dave&nbsp;Rand</A>&nbsp;
  <A HREF="mailto:dlr\@bungi.com">&lt;dlr\@bungi.com&gt;</A>
  <TR VALIGN=top>
  <TD WIDTH=88 ALIGN=RIGHT><FONT FACE="Arial,Helvetica" SIZE=2>$version</FONT></TD>
  <TD WIDTH=388 ALIGN=RIGHT><FONT FACE="Arial,Helvetica" SIZE=2>
  <A HREF="http://www.uni-ulm.de/~rbawidam/">Rainer&nbsp;Bawidamann</A>&nbsp;
  <A HREF="mailto:rb1\@informatik.uni-ulm.de">&lt;rb1\@informatik.uni-ulm.de&gt;</A></FONT>
  </TD>
</TR>
</TABLE>
EOT
$footer .= $q->end_html;

### the main switch
# the modes:
# if parameter "dir" is given show a list of the targets in this "directory"
# elsif parameter "png" is given show a graphic for the target given w/ parameter "log"
# elsif parameter "log" is given show the page for this target
# else show a list of directories and of targets w/o directory
# parameter "cfg" can hold the name of the config file to use
if (defined $CGI::dir) {
	# show a list of targets in the given directory
	my @httphead;
	if (yesorno($config{writeexpires})) {
		push @httphead, (-expires => '+' . int($config{interval}) . 'm');
	}
	if (yesorno($config{refresh})) {
		push @httphead, (-refresh => $config{refresh});
	}
	my @htmlhead = (-title => "MRTG/RRD - Group $CGI::dir", @headeropts);
	if ($targets{_}{addhead}) {
		push @htmlhead, (-head => "$targets{_}{addhead}");
	}
	print $q->header(@httphead), $q->start_html(@htmlhead);
	print $q->h1("Available Targets"),"\n\<table width=100\%>\n";
	my $cfgstr = (defined $CGI::cfg ? "&cfg=$CGI::cfg" : '');
	my $column = 0;
	foreach my $tar (@sorted) {
		my $small = 0;
		if (yesorno($options{$tar}{indexgraph})) {
			$small = $options{$tar}{indexgraph};
		}
		next if $tar =~ m/^[\$\^\_]$/; # _ is not a real target
		next if $targets{$tar}{directory} ne $CGI::dir;
		print '<tr>' if $column == 0;
		print '<td>',
			$q->p($q->a({href => "$meurl?log=$tar$cfgstr"}, "$targets{$tar}{title}"));
		print $q->a({href => "$meurl?log=$tar$cfgstr"},
			$q->img({src => "$meurl?log=$tar&png=$small&small=1$cfgstr",
				alt => "index-graph",
				getpngsize(getpngdir($tar)."$tar-$small.png")})
			) if $small;
		print "\</td>\n";
		$column++;
		if ($column >= $config{'columns'}) {
			$column = 0;
			print '</tr>';
		}
	}
	if ($column != 0 and $column < $config{'columns'}) {
		print '<td>&nbsp;</td>' x ($config{'columns'} - $column),"\</tr>\n";
	}
	print '</table>', $footer;
} elsif (defined $CGI::png) {
	# send a graphic, create it if necessary
	print_error("CGI call error") if (!defined $CGI::log);
	my ($start, $end, $maxage, $xs, $ys) = get_graph_params($CGI::log, $CGI::png);
	print_error("undefined graph") unless $start;

	my $rrd = getworkdir($CGI::log).$CGI::log.'.rrd';
	my $pngdir = getpngdir($CGI::log);
	my $pngfile = "${pngdir}$CGI::log-$CGI::png.png";
	
	# build the rrd command line: set the starttime and the graphics format (PNG)
	my @args = ($pngfile, '-s', $start, '-e', $end, '-a', 'PNG');
	# if it's not a small picture set the legends
	my ($l1,$l2,$l3,$l4,$li,$lo) = ('','','','','','');
	my ($ri, $ro) = ('','');

	push @args, '-w', $xs, '-h', $ys;
	if (!defined $CGI::small) {
		if ($targets{$CGI::log}{ylegend}) {
			push @args, '-v', $targets{$CGI::log}{ylegend}; }
		if ($targets{$CGI::log}{legend1}) {
			$l1 = ":".$targets{$CGI::log}{legend1}."\\l"; }
		if ($targets{$CGI::log}{legend2}) {
			$l2 = ":".$targets{$CGI::log}{legend2}."\\l"; }
		if ($targets{$CGI::log}{legend3} ne '--CALC--') {
			$l3 = ":".$targets{$CGI::log}{legend3}."\\l";
		} else {
			$l3 = ":Maximal 5 Minute ".$targets{$CGI::log}{legend1}."\\l";
		}
		if ($targets{$CGI::log}{legend4} ne '--CALC--') {
			$l4 = ":".$targets{$CGI::log}{legend4}."\\l";
		} else {
			$l4 = ":Maximal 5 Minute ".$targets{$CGI::log}{legend2}."\\l";
		}
		if ($targets{$CGI::log}{legendi}) {
			$li = $targets{$CGI::log}{legendi}; }
		else {	$li = "In: "; }
		$li =~ s':'\\:'; # ' quote :
		if ($targets{$CGI::log}{legendo}) {
			$lo = $targets{$CGI::log}{legendo}; }
		else {	$lo = "Out:"; }
		$lo =~ s':'\\:'; # ' quote :
		if ($options{$CGI::log}{integer}) {
			$li .= ' %9.0lf';
			$lo .= ' %9.0lf';
			$ri = '%3.0lf%%';
			$ro = '%3.0lf%%';
		} else {
			$li .= ' %8.3lf';
			$lo .= ' %8.3lf';
			$ri = '%6.2lf%%';
			$ro = '%6.2lf%%';
		}
		if ($targets{$CGI::log}{kmg}) {
			$li .= ' %s';
			$lo .= ' %s';
			if ($targets{$CGI::log}{kilo}) {
				push @args, '-b', $targets{$CGI::log}{kilo};
			}
			if ($targets{$CGI::log}{'shortlegend'}) {
				$li .= $targets{$CGI::log}{'shortlegend'};
				$lo .= $targets{$CGI::log}{'shortlegend'};
			}
		}
	}
	my $pngchar = substr($CGI::png,0,1);
	if ($pngchar and $targets{$CGI::log}{unscaled} and
		   $targets{$CGI::log}{unscaled} =~ m/$pngchar/) {
		my $max = intmax($targets{$CGI::log}{maxbytes},
			$targets{$CGI::log}{maxbytes1},
			$targets{$CGI::log}{maxbytes2},
			$targets{$CGI::log}{absmax});
		push @args, '-l', 0, '-u', $max, '-r';
	} elsif (yesorno($options{$CGI::log}{'logarithmic'})) {
		push @args, '-o';
	}
	push @args,'--alt-y-grid','--lazy','-c','MGRID#ee0000','-c','GRID#000000';
	my $factor = 1; # should we scale the values?
	if ($options{$CGI::log}{perminute}) {
		$factor = 60; # perminute -> 60x
	} elsif ($options{$CGI::log}{perhour}) {
		$factor = 3600; # perhour -> 3600x
	}
	if ($options{$CGI::log}{bits}) {
		$factor *= 8; # bits instead of bytes -> 8x
	}
	# now build the graph calculation commands
	# ds0/ds1 hold the normal data sources to graph/gprint
	my ($ds0, $ds1) = ('in', 'out');
	push @args, "DEF:$ds0=$rrd:ds0:AVERAGE", "DEF:$ds1=$rrd:ds1:AVERAGE";
	if (defined $options{$CGI::log}{unknaszero}) {
		push @args, "CDEF:uin=$ds0,UN,0,$ds0,IF",
			"CDEF:uout=$ds1,UN,0,$ds1,IF";
		($ds0, $ds1) = ('uin', 'uout');
	}
	if ($factor > 1) {
		# scale the values. we need a CDEF for this
		push @args, "CDEF:fin=$ds0,$factor,*","CDEF:fout=$ds1,$factor,*";
		($ds0, $ds1) = ('fin', 'fout');
	}
	my $maximum0 = $targets{$CGI::log}{maxbytes1};
	$maximum0 = $targets{$CGI::log}{maxbytes} unless $maximum0;
	$maximum0 = 1 unless $maximum0;
	my $maximum1 = $targets{$CGI::log}{maxbytes2};
	$maximum1 = $targets{$CGI::log}{maxbytes} unless $maximum1;
	$maximum1 = 1 unless $maximum1;
	# ps0/ps1 hold the percentage data source for gprint
	my ($ps0, $ps1) = ('pin', 'pout');
	push @args, "CDEF:pin=$ds0,$maximum0,/,100,*,$factor,/",
		"CDEF:pout=$ds1,$maximum1,/,100,*,$factor,/";
	# the commands to draw the values
	push @args, "AREA:$ds0#00cc00$l1", "LINE1:$ds1#0000ff$l2";
	if (yesorno($options{$CGI::log}{'graphtotal'})) {
		push @args, "CDEF:total=$ds0,$ds1,+", "LINE1:total#ffa050:Total AVG\\l";
	}
	
	# now for the peak graphs / maximum values
	# mx0/mx1 hold the maximum data source for graph/gprint
	my ($mx0, $mx1) = ($ds0, $ds1);
	# px0/px1 hold the maximum pecentage data source for gprint
	my ($px0, $px1) = ($ps0, $ps1);
	if (!defined $CGI::small) {
		# the defs for the maximum values: for the legend ('MAX') and probabely
		# for the 'withpeak' graphs
		push @args, "DEF:min=$rrd:ds0:MAX", "DEF:mout=$rrd:ds1:MAX";
		($mx0, $mx1) = ('min', 'mout');
		if (defined $options{$CGI::log}{unknaszero}) {
			push @args, "CDEF:umin=$mx0,UN,0,$mx0,IF",
				"CDEF:umout=$mx1,UN,0,$mx1,IF";
			($ds0, $ds1) = ('umin', 'umout');
		}
		if ($factor > 1) {
			# scale the values. we need a CDEF for this
			push @args, "CDEF:fmin=$mx0,$factor,*","CDEF:fmout=$mx1,$factor,*";
			($mx0, $mx1) = ('fmin', 'fmout');
		}
		# draw peak lines if configured
		if (substr($CGI::png,0,1) =~ /[$targets{$CGI::log}{withpeak} ]/) {
			push @args, "LINE1:$mx0#006600$l3", "LINE1:$mx1#ff00ff$l4";
			push @args, "CDEF:pmin=$mx0,$maximum0,/,100,*,$factor,/",
				"CDEF:pmout=$mx1,$maximum1,/,100,*,$factor,/";
			($px0, $px1) = ('pmin', 'pmout');
			if (yesorno($options{$CGI::log}{'graphtotal'})) {
				push @args, "CDEF:mtotal=$mx0,$mx1,+", "LINE1:mtotal#ff5050:Total MAX\\l";
			}
		}
	}
	if (!defined $CGI::small) {
		# print the legends
		if ($options{$CGI::log}{nopercent}) {
			push @args,
				"GPRINT:$mx0:MAX:Maximal $li",
				"GPRINT:$mx1:MAX:Maximal $lo\\l",
				"GPRINT:$ds0:AVERAGE:Average $li",
				"GPRINT:$ds1:AVERAGE:Average $lo\\l",
				"GPRINT:$ds0:LAST:Current $li",
				"GPRINT:$ds1:LAST:Current $lo\\l";
		} else {
			push @args,
				"GPRINT:$mx0:MAX:Maximal $li",
				"GPRINT:$px0:MAX:($ri)",
				"GPRINT:$mx1:MAX:Maximal $lo",
				"GPRINT:$px1:MAX:($ro)\\l",
				"GPRINT:$ds0:AVERAGE:Average $li",
				"GPRINT:$ps0:AVERAGE:($ri)",
				"GPRINT:$ds1:AVERAGE:Average $lo",
				"GPRINT:$ps1:AVERAGE:($ro)\\l",
				"GPRINT:$ds0:LAST:Current $li",
				"GPRINT:$ps0:LAST:($ri)",
				"GPRINT:$ds1:LAST:Current $lo",
				"GPRINT:$ps1:LAST:($ro)";
		}
	}
	# fire up rrdtool
	my ($a, $rrdx, $rrdy) = RRDs::graph(@args);
	my $e = RRDs::error();
	# if the file given to rrdtool is not a png file rrdtool won't report
	# an error (fixed in 1.0.11). we have to check the reported image size:
	if ($e or $rrdx == 0 and $rrdy == 0) {
		senderrorpng($xs, $ys, "cannot write to graph dir $pngdir"
			.($e ? "\nrrdtool error: $e" : '')) if ! -w $pngdir;
		senderrorpng($xs, $ys, "cannot write $pngfile"
				.($e ? "\nrrdtool error: $e" : "\nadditional info: file is not a PNG file"))
			if (-e $pngfile and !-w _);
		unlink $pngfile or senderrorpng($xs, $ys, "cannot delete file $pngfile")
			if -e $pngfile; # do we get here if it does not exist?
		($a, $rrdx, $rrdy) = RRDs::graph(@args);
		$e = RRDs::error();
	}
	if ($e) {
		senderrorpng($xs, $ys, "RRDTool error: $e");
	} elsif ($rrdx == 0 and $rrdy == 0) {
		senderrorpng($xs, $ys, "Cannot create graph or wrong format")
			or print_error("Cannot create graph or wrong format");
	}
	# no error, try to send the file
	open(PNG, "<$pngfile")
		or senderrorpng($xs, $ys, "cannot read graph file")
		or print_error("cannot read graph file");
	my @httphead = (-type => "image/png");
	if (yesorno($config{writeexpires})) {
		push @httphead, (-expires => "+${maxage}s");
	}
	print $q->header(@httphead);
	binmode(PNG); binmode(STDOUT);
	while(read PNG, my $buf,4096) { print STDOUT $buf; }
	close PNG;
} elsif (defined $CGI::log) {
	# show the graphics for one target
	print_error ("Target $CGI::log unknown") if (!exists $targets{$CGI::log});
	my $title;
	# user defined title?
	if ($targets{$CGI::log}{title}) {
		$title = $targets{$CGI::log}{title};
	} else {
		$title = "MRTG/RRD - Target $CGI::log";
	}
	my @httphead;
	if (yesorno($config{writeexpires})) {
		push @httphead, (-expires => '+' . int($config{interval}) . 'm');
	}
	if (yesorno($config{refresh})) {
		push @httphead, (-refresh => $config{refresh});
	}
	my @htmlhead = (-title => $title, @headeropts);
	if ($targets{$CGI::log}{addhead}) {
		push @htmlhead, (-head => "$targets{$CGI::log}{addhead}");
	}
	print $q->header(@httphead), $q->start_html(@htmlhead);
	# user defined header line? (should exist as mrtg requires it)
	if (exists $targets{$CGI::log}{pagetop}) {
		print $targets{$CGI::log}{pagetop},"\n";
	} else {
		print $q->h1("Target $targets{$CGI::log}{title}"),"\n";
	}
	my $lasttime = RRDs::last("$config{workdir}/$targets{$CGI::log}{directory}/${CGI::log}.rrd");
	print $q->hr,
		"The statistics were last updated: ",$q->b(scalar(localtime($lasttime))),
		$q->hr if $lasttime;
	my $sup = $targets{$CGI::log}{suppress};
	my $url = "$meurl?log=$CGI::log";
	$url .= "&cfg=$CGI::cfg" if defined $CGI::cfg;
	$url .= "&png";
	# the header lines and tags for the graphics
	if ($sup !~ /d/) {
		print $q->h2("'Daily' graph (5 Minute Average)"),"\n",
			$q->img({src => "$url=daily", alt => "daily-graph",
				getpngsize(getpngdir($CGI::log)."$CGI::log-daily.png")}
			), "\n";
	}
	if ($sup !~ /w/) {
		print $q->h2("'Weekly' graph (30 Minute Average)"),"\n",
			$q->img({src => "$url=weekly", alt => "weekly-graph",
				getpngsize(getpngdir($CGI::log)."$CGI::log-weekly.png")}
			), "\n";
	}
	if ($sup !~ /m/) {
		print $q->h2("'Monthly' graph (2 Hour Average)"),"\n",
			$q->img({src => "$url=monthly", alt => "monthly-graph",
				getpngsize(getpngdir($CGI::log)."$CGI::log-monthly.png")}
			), "\n";
	}
	if ($sup !~ /y/) {
		print $q->h2("'Yearly' graph (1 Day Average)"),"\n",
			$q->img({src => "$url=yearly", alt => "yearly-graph",
				getpngsize(getpngdir($CGI::log)."$CGI::log-yearly.png")}
			), "\n";
	}
	if ($targets{$CGI::log}{pagefoot}) {
		print $targets{$CGI::log}{pagefoot};
	}
	print $footer;
} else {
	# no parameter - show a list of directories and targets without "Directory[...]" (aka root-targets)
	my @httphead;
	if (yesorno($config{writeexpires})) {
		push @httphead, (-expires => '+1d'); # how often do you add targets?
	}
	if (yesorno($config{refresh})) {
		push @httphead, (-refresh => $config{refresh});
	}
	my @htmlhead = (-title => "MRTG/RRD $version", @headeropts);
	if ($targets{_}{addhead}) {
		push @htmlhead, (-head => "$targets{_}{addhead}");
	}
	print $q->header(@httphead), $q->start_html(@htmlhead);
	my (@dirs, %dirs, @logs);
	# get the list of directories and "root"-targets
	foreach my $tar (@sorted) {
		next if $tar =~ m/^[_\$\^]$/; # pseudo targets
		if ($targets{$tar}{directory}) {
			next if exists $dirs{$targets{$tar}{directory}};
			$dirs{$targets{$tar}{directory}} = $tar;
			push @dirs, $targets{$tar}{directory};
		} else {
			push @logs, $tar;
		}
	}
	my $cfgstr = (defined $CGI::cfg ? "&cfg=$CGI::cfg" : '');
	print $q->h1("Available Targets"),"\n";
	if ($#dirs > -1) {
		print $q->h2("Directories"),"\n\<table width=100\%>\n";
		my $column = 0;
		foreach my $tar (@dirs) {
			print '<tr>' if $column == 0;
			print $q->td($q->a({href => "$meurl?dir=$tar$cfgstr"},
				"Group $tar")),"\n";
			$column++;
			if ($column >= $config{'columns'}) {
				$column = 0;
				print '</tr>';
			}
		}
		if ($column != 0 and $column < $config{'columns'}) {
			print '<td>&nbsp;</td>' x ($config{'columns'} - $column),"\</tr>\n";
		}
		print '</table><hr>';
	}
	if ($#logs > -1) {
		print $q->h2("Targets"),"\n\<table width=100\%>\n";
		my $column = 0;
		foreach my $tar (@logs) {
			my $small = 0;
			if (yesorno($options{$tar}{indexgraph})) {
				$small = $options{$tar}{indexgraph};
			}
			next if $tar =~ m/^[\$\^_]$/;
			next if $targets{$tar}{directory};
			print '<tr>' if $column == 0;
			print '<td>',
				$q->p($q->a({href => "$meurl?log=$tar$cfgstr"},"$targets{$tar}{title}"));
			print $q->a({href => "$meurl?log=$tar$cfgstr"},
				$q->img({src => "$meurl?log=$tar&png=$small&small=1$cfgstr",
					alt => "index-graph",
					getpngsize(getpngdir($tar)."$tar-$small.png")}))
				if $small;
			print "\</td>\n";
			$column++;
			if ($column >= $config{'columns'}) {
				$column = 0;
				print '</tr>';
			}
		}
		if ($column != 0 and $column < $config{'columns'}) {
			print '<td>&nbsp;</td>' x ($config{'columns'} - $column),"\</tr>\n";
		}
		print '</table>';
	}
	print $footer;
}
exit 0;

sub build_value($$) {
	my ($opt, $cval) = @_;
	my $val = ($targets{'^'}{$opt} ? $targets{'^'}{$opt}.' ' : '')
		. $cval
		. ($targets{'$'}{$opt} ? ' '.$targets{'$'}{$opt} : '');
	$val =~ s/^\s*(.*?)\s*$/$1/;
	return $val;
}

# read the mrtg.cfg file
sub read_mrtg_config()
{
	my ($opt, $tar, $val);
	my $line = '';
	my @lines;
	open(CFG, "<$cfgfile") || print_error("Cannot open config file");
	while(<CFG>) {
		chomp;		# remove newline
		s/\s+$//g;	# remove trailing space
		s/\s/ /g;	# collapse white space to one space
		next if /^ *\#/;# ignore comment lines
		next if /^ *$/;	# ignore empty lines
		if (/^ +(.*)$/) {	# continuation lines
			$lines[$#lines] .= "\n".$1;
		} else {
			push @lines, $_;
		}
	}
	close CFG;
	# set some defaults
	my %defaults = (
		directory => '',
		suppress => '',
		interval => 5,
		xsize => 400,
		ysize => 100,
		withpeak => '',
		ylegend => 'Bytes per Second',
		legend1 => 'Incoming Traffic in Bytes per Second',
		legend2 => 'Outgoing Traffic in Bytes per Second',
		legend3 => '--CALC--',
		legend4 => '--CALC--',
		legendi => 'In: ',
		legendo => 'Out:',
		shortlegend => '--CALC--',
		kmg => ',k,M,G,T,P',
		kilo => 1000,
		);
	my %defoptions = (
		'indexgraph' => 'daily.s',
		'graph daily.s' => '-1250m   now    300',
		'graphsize daily.s' => '250 100',
		# don't set a default graphsize so override rules work
		# 'graphsize' => '400 100',
		'graph daily' => '-2000m   now    300',
		'graph weekly' => '-12000m  now   1800',
		'graph monthly' => '-800h    now   7200',
		'graph yearly' => '-400d    now  86400',
		'graphtotal' => 'no',
		'logarithmic' => 'no',
		'error pic' => undef,
		);
	%config = (
		writeexpires => 'yes',
		background => '#ffffff',
		interval => 5,
		icondir => '',
		columns => 1,
		refresh => 'yes',
		);
	%{$options{_}} = %defoptions;
	%{$targets{_}} = %defaults;
	%{$targets{'$'}} = (); # prepend and append 'targets' are empty by default
	%{$targets{'^'}} = ();
	foreach (@lines) {
		if (/^([\d\w]+)\[(\S*)\] *: *(.*)$/s) {
			# a target config line
			($tar, $opt, $val) = (lc($2), lc($1), $3);
			$val = '' if !defined $val; # get around undef values
			if (!exists $targets{$tar}) {
				# first occurance of a target, copy defaults
				# (don't need to check for '_' as this exists in any case)
				push @sorted, $tar;
				foreach my $k (keys %{$targets{_}}) {
					$targets{$tar}{$k} = build_value($k, $targets{_}{$k});
				}
				# copy prefix-only settings
				foreach my $k (keys %{$targets{'^'}}) {
					$targets{$tar}{$k} = build_value($k, '')
						unless $targets{$tar}{$k};
				}
				# copy postfix-only settings. no need for build_value here
				foreach my $k (keys %{$targets{'$'}}) {
					$targets{$tar}{$k} = $targets{'$'}{$k}
						unless $targets{$tar}{$k};
				}
				%{$options{$tar}} = %{$options{_}};
				# copy options from ^ and $ (thx to Mike Fisher)
				map {$options{$tar}{$_} = 1}
					(keys %{$options{'^'}}, keys %{$options{'$'}});
				# set a default title if none given
				# (shouldn't happen as mrtg requires a title)
				$targets{$tar}{'title'} = $tar unless $targets{$tar}{'title'};
			}
			if ($tar eq '_' && $val eq '') {
				# a line "Command[_]:", resets default
				if ($defaults{$opt}) {
					# there is a default for this, set it
					$targets{'_'}{$opt} = $defaults{$opt};
				} else {
					# no default, delete from _
					delete $targets{'_'}{$opt};
				}
			} elsif ($opt eq 'options') {
				# "Options[...]: " - add values to a set of options
				# not sure about this: mrtg allows only one "Options" line,
				# so the defaults should probably be replaced not merged
				$val = lc($val);
				map {$options{$tar}{$_} = 1} split(/[,\s]+/, $val);
			} elsif ($tar eq '$' || $tar eq '^') {
				$targets{$tar}{$opt} = $val;
			} else {
				# "Command[...]: ..."
				$targets{$tar}{$opt} = build_value($opt, $val);
			}
			next;
		} elsif (/^([\d\w]+) *: *(.*)$/s) {
			# global option
			($tar, $opt, $val) = (undef, lc($1), $2);
			$config{$opt} = $val;
			next;
		}
		$_ =~ s/</&gt;/g;
		print STDERR scalar(localtime())," MRTG/RRD 14all.cgi: unknown field in mrtg.conf '$_'\n";
	}
	foreach $tar (keys %targets) {
		next if $tar eq '_';
		# we need to replace '&nbsp;' in the legends as we print them via rrdtool
		foreach $opt (qw/legend1 legend2 legend3 legend4 legend5 legendi legendo ylegend shortlegend/) {
			if (exists $targets{$tar}{$opt}) {
				$targets{$tar}{$opt} =~ s'&nbsp;' ';   #'
				$targets{$tar}{$opt} =~ tr/\n/ /;
				# escape % in legendi,legendo and shortlegend
				$targets{$tar}{$opt} =~ s/%/%%/g if $opt =~ m/[io]/;
				if ($options{$tar}{'bits'} and $targets{$tar}{$opt} eq $defaults{$opt}) {
					$targets{$tar}{$opt} =~ s/Bytes/Bits/g;
				}
			}
		}
		# add settings from userrdtool[...] to options
		if (exists $targets{$tar}{'userrdtool'}) {
			foreach my $line (split(/\n/, $targets{$tar}{'userrdtool'})) {
				if (index($line, ':') > 0) {
					# line contains a ':'
					($opt, $val) = ($line =~ m/^([^,]+) *: *(.*)$/);
					$options{$tar}{$opt} = $val;
				} else {
					# no ':' just notice it's there
					$options{$tar}{$line} = 1;
				}
			}
		}
		if (defined $targets{$tar}{'shortlegend'} and 
				$targets{$tar}{'shortlegend'} eq '--CALC--') {
			$targets{$tar}{'shortlegend'} = ($options{$tar}{'bits'} ? 'b/' :'B/') .
				($options{$tar}{'perhour'} ? 'h' : $options{$tar}{'perminute'} ? 'min' : 's');
		}
	}
	# add settings from userrdtool to config
	if (exists $config{'userrdtool'}) {
		foreach my $line (split(/\n/, $config{'userrdtool'})) {
			if (index($line, ':') > 0) {
				# line contains a ':'
				($opt, $val) = ($line =~ m/^([^:]+) *: *(.*)$/);
				$config{$opt} = $val;
			} else {
				# no ':' just notice it's there
				$config{$line} = 1;
			}
		}
	}
	if (yesorno($config{refresh}) and $config{refresh} !~ m/^\d*[1-9]\d*$/) {
		$config{refresh} = $config{interval} * 60;
	}
}

sub print_error(@)
{
	print $q->header, $q->start_html(
		-title => 'MRTG/RRD index.cgi - Skript error',
		@author,
		-bgcolor => "#ffffff"),
		$q->h1('Skript Error'),
		@_, $q->end_html;
	exit 0;
}

sub intmax(@)
{
	my (@p) = @_;
	my $max = 0;
	foreach my $n (@p) {
		$max = int($n) if defined $n and int($n) > $max;
	}
	return $max;
}

sub yesorno($)
{
	my $opt = shift;
	return 0 unless defined $opt;
	return 0 if $opt =~ /^((no?)|(false)|0)$/i;
	return 1;
}

sub get_graph_params($$)
{
	my ($tar, $graph) = @_;
	my ($start, $end, $maxage, $xs, $ys);
	if ($options{$tar}{"graph $graph"}) {
		($start, $end, $maxage) = split / +/, $options{$tar}{"graph $graph"};
	} elsif ($options{_}{"graph $graph"}) {
		($start, $end, $maxage) = split / +/, $options{_}{"graph $graph"};
	} else {
		return ();
	}
	if ($options{$tar}{"graphsize $graph"}) {
		($xs, $ys) = split / +/, $options{$tar}{"graphsize $graph"};
	} elsif ($options{$tar}{"graphsize"}) {
		($xs, $ys) = split / +/, $options{$tar}{"graphsize"};
	} elsif ($options{_}{"graphsize $graph"}) {
		($xs, $ys) = split / +/, $options{_}{"graphsize $graph"};
	} elsif ($options{_}{"graphsize"}) {
		($xs, $ys) = split / +/, $options{_}{"graphsize"};
	} else {
		$xs = $targets{$tar}{xsize};
		$ys = $targets{$tar}{ysize};
	}
	return ($start, $end, $maxage, $xs, $ys);
}

sub getworkdir($)
{
	my ($tar) = @_;
	my $workdir = "$config{workdir}";
	$workdir .= "/" unless $workdir =~ m|/$|;
	$workdir .= $targets{$tar}{directory};
	$workdir .= "/" unless $workdir =~ m|/$|;
	return $workdir;
}

sub getpngdir($)
{
	my ($tar) = @_;
	my $pngdir = getworkdir($tar);
	if (!-w $pngdir) {
		if ($^O =~ m/Win/i) {
			$pngdir = $ENV{'TEMP'};
			$pngdir = $ENV{'TMP'} unless $pngdir;
			$pngdir = '' unless $pngdir;
			$pngdir .= '/' unless $pngdir =~ m|/$|;
		} else { 
			$pngdir = '/tmp/';
		}
	}
	return $pngdir;
}

use IO::File;

sub pngstring() { return chr(137)."PNG".chr(13).chr(10).chr(26).chr(10); };

sub getpngsize($)
{
	my ($file) = @_;
	my $fh = new IO::File $file;
	return () unless defined $fh;
	my $line;
	if (sysread($fh, $line, 8) != 8 or $line ne pngstring()) {
		$fh->close;
		return ();
	}
	CHUNKS: while(1) {
		last CHUNKS if (sysread($fh, $line, 8) != 8);
		my ($chunksize, $type) = unpack "Na4", $line;
		if ($type ne "IHDR") {
			last CHUNKS if (sysread($fh, $line, $chunksize + 4) != $chunksize + 4);
			next CHUNKS;
		}
		last CHUNKS if (sysread($fh, $line, 8) != 8);
		$fh->close;
		my ($x, $y) = unpack("NN", $line);
		return ('-width' => "$x", '-height' => "$y");
	}
	$fh->close;
	return ();
}

# this data contains a small png with the text:
# "error: cannot create graph"

sub errorpng()
{
	return (
		137,80,78,71,13,10,26,10,0,0,0,13,73,72,68,82,0,0,0,187,0,0,0,29,4,3,0,
		0,0,0,251,0,170,0,0,0,4,103,65,77,65,0,0,177,143,11,252,97,5,0,0,0,30,80,
		76,84,69,255,0,0,255,93,93,255,128,128,255,155,155,255,176,176,255,195,
		195,255,212,212,255,227,227,255,241,241,255,255,255,17,191,146,253,0,0,
		0,56,116,69,88,116,83,111,102,116,119,97,114,101,0,88,86,32,86,101,114,
		115,105,111,110,32,51,46,49,48,97,32,32,82,101,118,58,32,49,50,47,50,57,
		47,57,52,32,40,80,78,71,32,112,97,116,99,104,32,49,46,50,41,221,21,46,73,
		0,0,2,40,73,68,65,84,120,218,237,147,177,107,219,64,20,198,159,34,233,92,
		109,130,180,132,108,55,180,78,189,169,113,8,220,166,80,106,208,230,102,
		48,237,118,56,246,153,219,28,7,2,183,165,77,23,109,142,101,157,244,254,
		219,62,73,198,113,106,211,150,144,108,254,166,143,167,79,63,221,125,167,
		3,216,107,175,189,94,65,136,56,223,26,58,166,224,207,71,6,114,3,175,148,
		220,10,136,66,47,183,134,44,254,115,114,48,252,55,126,87,192,92,248,24,
		254,237,173,70,110,246,95,120,182,184,17,231,242,4,103,160,79,34,213,12,
		117,148,224,204,205,211,159,96,205,18,24,254,162,26,105,189,198,66,29,149,
		96,80,10,172,29,64,154,91,55,83,30,22,92,124,43,33,152,60,86,75,229,196,
		172,204,68,42,83,85,114,157,70,4,97,182,218,121,121,150,187,56,41,0,111,
		75,254,253,54,141,197,13,64,203,154,184,138,78,114,223,158,47,222,229,156,
		28,128,95,104,235,150,182,115,159,244,133,53,211,160,208,215,27,71,43,153,
		253,36,238,223,23,96,250,250,42,84,43,188,55,58,180,174,117,16,202,208,
		156,166,97,114,87,237,185,211,239,72,138,230,78,217,30,126,200,220,204,
		35,199,225,227,67,139,240,167,189,11,33,197,244,120,17,60,180,178,39,229,
		60,128,136,217,156,74,211,117,227,53,30,18,164,29,211,115,11,186,155,1,
		155,87,120,93,157,123,204,104,77,145,65,194,215,14,68,228,90,119,9,30,173,
		148,108,22,72,246,20,63,167,4,91,64,71,234,213,48,116,204,89,193,31,241,
		57,28,173,240,46,167,104,214,237,30,229,135,21,158,92,8,162,239,85,209,
		100,22,72,17,251,187,241,94,93,14,64,151,6,250,171,143,108,225,173,241,
		235,114,146,233,177,172,162,142,13,174,91,85,57,228,160,41,39,3,205,147,
		166,156,77,60,98,86,227,65,163,13,117,243,189,35,196,31,30,150,107,124,
		36,176,228,111,170,67,71,12,235,232,172,133,101,118,128,146,28,253,160,
		77,52,193,84,138,20,227,77,188,82,106,232,73,104,115,120,171,46,160,71,
		3,26,58,131,49,135,193,229,136,238,141,130,17,244,184,171,46,193,165,39,
		78,239,170,142,142,67,71,125,30,194,32,38,71,249,193,200,82,212,31,183,
		99,241,101,76,247,207,219,125,223,158,41,49,111,126,134,202,70,47,9,110,
		228,151,230,238,21,241,123,237,245,162,250,13,181,158,203,16,233,3,210,
		153,0,0,0,7,116,73,77,69,7,208,1,19,13,28,15,223,54,180,209,0,0,0,0,73,
		69,78,68,174,66,96,130
	);
}

sub senderrorpng($$$) {
	my ($xs, $ys, $text) = @_;
	if (defined $options{$CGI::log}{'error pic'}) {
		open(PNG, $options{$CGI::log}{'error pic'}) || return 0;
		print $q->header(-type => "image/png");
		binmode(PNG); binmode(STDOUT);
		while(read PNG, my $buf,4096) { print STDOUT $buf; }
		close PNG;
		exit 0;
	}
	if ($DEBUG) {
		eval { require GD; 1; };
		unless ($@) {
			my $im = new GD::Image($xs, $ys);
			my $back = $im->colorAllocate(255,255,255);
			$im->transparent($back);
			my $red = $im->colorAllocate(255,0,0);
			$im->filledRectangle(0,0,$xs-1,$ys-1,$back);
			my @textsplit = split(/\n/, $text);
			my $starty = 10;
			foreach $text (@textsplit) {
				$im->string(GD::gdMediumBoldFont(), 10, $starty, $text, $red);
				$starty += 5 + GD::gdMediumBoldFont()->height;
			}
			binmode(STDOUT);
			if ($GD::VERSION lt '1.20') {
				print $q->header(-type => "image/gif");
				eval 'print $im->gif';
			} elsif ($GD::VERSION ge '1.20') {
				print $q->header(-type => "image/png");
				eval 'print $im->png';
			}
			exit 0;
		}
	}
	print $q->header(-type => "image/png");
	binmode(STDOUT);
	print pack("C*", errorpng());
	exit 0;
}
